"""
Module to assist in the super-common operation of
joining values in sequences into strings.
"""

import re
from options import Options
from .util import *
from .quoter import Quoter
import six


class Joiner(Quoter):

    """
    A type of Quoter that deals with sequences.
    """

    styles = {}         # remember named styles

    options = Quoter.options.add(
        sep=', ',    # separator between items
        twosep=None,   # separator between items if only two
        lastsep=None,  # separator between penultimate and final item
        quoter=None,   # quoter for individual items
        endcaps=None,  # quoter for entire joined sequence
    )

    def __init__(self, *args, **kwargs):
        """
        Create a Joiner
        """
        # Restating basic init to avoid errors of self.__getattribute__
        # that can flummox superclass instantiation
        Quoter.__init__(self)

        opts = self.options = self.__class__.options.push(kwargs)
        self._register_name(opts.name)

    def __call__(self, seq, **kwargs):
        """
        Join the items of a sequence into a string. Implicitly stringifies any
        not-string values. Allows specification of the separator between items (and
        a special case for the last separator). Allows each item to be optionally
        quoted by a function, and the entire list to be optionally quoted with an
        endcaps function. A separate suffix and prefix may also be provdied.
        """
        opts = self.options.push(kwargs)

        def prep(v):
               """
               Prepare an item by stringifying and optionally quoting it.
               """
               s = stringify(v)
               return opts.quoter(s) if opts.quoter else s

        seqlist = list(seq)
        length = len(seqlist)
        if length == 0:
            core = ''
        elif length == 1:
            core = prep(seqlist[0])
        elif length == 2 and opts.twosep:
            sep = opts.twosep if opts.twosep is not None else opts.sep
            core = sep.join(prep(v) for v in seqlist)
        else:
            start = [ prep(v) for v in seqlist[:-1] ]
            final = prep(seqlist[-1])
            if opts.lastsep is None:
                opts.lastsep = opts.sep
            core = opts.lastsep.join([ opts.sep.join(start), final])
        pstr, mstr = self._whitespace(opts)
        capped = opts.endcaps(core) if opts.endcaps else core
        payload = [mstr, blanknone(opts.prefix), pstr, capped, pstr,
                         blanknone(opts.suffix), mstr]
        return self._output(payload, opts)


# FIXME: issue with named styles that have multiple styles extant

join = Joiner()

# specializations

# A and B. A, B, and C.
and_join = join.but(sep=', ', twosep=' and ', lastsep=', and ', name='and_join and')
word_join = and_join # deprecated

# A or B. A, B, or C.
or_join = join.but(sep=', ', twosep=' or ', lastsep=', or ', name='or_join or')

joinlines = join.but(sep="\n", suffix="\n", name='joinlines lines')

# TODO: Rationalize with respect to more sophisticated quoter args
# TODO: Rationalizw wrt more sophisticated quoter class structure and extensibility
# TODO: Add padding and margin, like quoter

concat = join.but(sep='', twosep='', lastsep='', name='concat')


def is_sequence(arg):
    """
    Is a list, set etc. Not a string.
    """
    if hasattr(arg, "__iter__") or hasattr(arg, "__getitem__"):
        if not hasattr(arg, "strip"):
            return True
    return False


items_options = Options(
    sep="\n",  # separator between items
    fmt="{key}: {value!r}",
    header=None,   # header for entire list
    footer=None    # footer for entire list
)


def iter_items(items):
    if hasattr(items, 'items'):  # dict or mapping
        for k, v in items.items():
            yield k, v
    else:
        for k, v in enumerate(items):
            yield k, v


def items(seq, **kwargs):
    opts = items_options.push(kwargs)

    formatted_items = [ opts.fmt.format(key=k, value=v) for k,v in iter_items(seq) ]
    items_str = opts.sep.join(formatted_items)
    if opts.header or opts.footer:
        parts = []
        if opts.header:
            parts.extend([opts.header, opts.sep])
        parts.append(items_str)
        if opts.footer:
            parts.extend([opts.sep, opts.footer])
        items_str = ''.join(parts)
    return items_str

# TODO: needs to be moved into object struture, like quoter




import sys
import re
from itertools import chain

_PY3 = sys.version_info[0] > 2
if _PY3:
    basestring = unicode = str


def is_string(v):
    return isinstance(v, basestring)


def stringify(v):
    """
    Return a string. If it's already a string, just return that. Otherwise,
    stringify it. Under Python 3, this makes perfect sense. Under Python 2,
    if the string contains Unicode octets (e.g. UTF-8 bytes, because it's
    really a byte string pretending to be a full string), casting to unicode
    isn't safe. Solution: Use only for Unicode strings.
    """
    return v if isinstance(v, basestring) else unicode(v)


def blanknone(v):
    """
    Return a value, or empty string if it's None.
    """
    return '' if v is None else v


def pad(seq, n, fillvalue=None):
    """
    You want a sequence at least n items long, say for an unpacking
    operation. You might not have n items. This will right-pad your
    sequence into one of the desired length, using your
    desired fillvalue (by default, None).
    """

    length = len(seq)
    if length == n:
        return seq
    else:
        needed = n - length
        return seq + type(seq)([fillvalue] * needed)

    # itertools equivalent:
    # return chain(iterable, repeat(fillvalue, n-len(iterable)))


def parse_css_selector(spec):
    """
    Parse a CSS-style selector (similar to what jQuery and CSS use).
    Returns an dict with tag_, id, and class, and other key, value
    fields.
    """
    if spec is None:
        return {}

    # use regular expressions to parse the selector spec
    tagnames = re.findall(r'^(\w+)',      spec)
    classes  = re.findall(r'\.([\w\-]+)', spec)
    ids      = re.findall(r'\#([\w\-]+)', spec)
    attrs    = re.findall(r'\[([\w-]+=[^]]*)\]', spec)

    # make some basic assertions
    assert len(tagnames) <= 1
    assert len(ids) <= 1

    # construct the result dict
    atts = {}
    if ids:
        atts['id'] = ids[0]
    if classes:
        atts['class'] = ' '.join(classes)
    if tagnames:
        atts['_tag'] = tagnames[0]
    for a in attrs:
        k, v = a.split('=')
        atts[k] = v
    return atts


def style_attribs(spec):
    """
    If spec is a dictionary, assume it's the one we want. Else,
    parse it.
    """
    if spec is None:
        return {}
    elif isinstance(spec, dict):
        return spec
    else:
        return parse_css_selector(spec)


def combine_classes(c1=None, c2=None):
    c1list = (c1 or '').strip().split()
    c2list = (c2 or '').strip().split()

    if not (c1list or c2list):
        return None
    classes = []
    seen = set()
    for c in chain(c1list, c2list):
        if c not in seen:
            classes.append(c)
            seen.add(c)
    return ' '.join(classes)


def update_style_dict(d, other, reverse_class=True):
    """
    Update one style dict d, with the other style dict.
    Very like a dictionary update, except that the class attribute
    can be multi-valued and additive, not replaced.

    reverse_class, if True, means to process the other's
    class value (if any) before the receiving dictionary's.
    This may seem an odd, even wrong, choice. It's actually
    the right choice in the case one is applying a series of
    updates and wishes the last update to be the higest
    priority, over-rides all others update (but still have its
    class names to appear first in the list). If you don't like
    this choice, set it to False and no harm done. Either way
    is proper CSS.
    """
    if other is None or len(other) == 0: # nothing to merge; quick exit
        return
    dclass = d.get('class')
    oclass = other.get('class')
    d.update(other)
    if dclass or oclass:
        clslist = [oclass, dclass] if reverse_class else [dclass, oclass]
        d['class'] = combine_classes(*clslist)




from quoter import *
from quoter.joiner import is_sequence
import pytest
import six
import sys


### Helper functions

def single(s):
    return "'" + s + "'"


def brackets(s):
    return "[" + s + "]"


### Tests


def test_doc_example():

    mylist = list("ABCD")
    assert join(mylist) == "A, B, C, D"
    assert join(mylist, sep=" | ", endcaps=braces) == "{A | B | C | D}"
    assert join(mylist, sep=" | ", endcaps=braces.clone(padding=1)) == \
            "{ A | B | C | D }"
    assert join(mylist, sep=" | ", endcaps=braces.but(padding=1)) == \
            "{ A | B | C | D }"
    assert and_join(mylist[:2]) == "A and B"
    assert and_join(mylist[:3]) == "A, B, and C"
    assert and_join(mylist) == "A, B, C, and D"
    assert and_join(mylist, quoter=double, lastsep=" and ") == \
            '"A", "B", "C" and "D"'


def test_margins_and_padding():
    mylist = list("ABCD")
    bl = join.but(sep="|", prefix='{', suffix='}')

    assert bl(mylist) =="{A|B|C|D}"
    assert bl(mylist, margin=1) ==" {A|B|C|D} "
    assert bl(mylist, padding=2) =="{  A|B|C|D  }"
    assert bl(mylist, margin=2, padding=1) =="  { A|B|C|D }  "


def test_is_sequence():
    assert not is_sequence('this')
    assert not is_sequence(six.u('this'))
    assert not is_sequence(4)
    assert not is_sequence(4.4)
    assert not is_sequence(3.3+2j)
    assert is_sequence([])
    assert is_sequence([1,3,4])
    assert is_sequence(set())
    assert is_sequence(set([1,3,4]))
    assert is_sequence(range(10))


def test_oxford():
    assert word_join([]) == ''
    assert word_join([1]) == '1'
    assert word_join([1, 2]) == '1 and 2'
    assert word_join([1, 2, 3]) == '1, 2, and 3'
    assert word_join([1, 2, 3, 4]) == '1, 2, 3, and 4'


def test_heathen():
    # non Oxford Comma
    assert word_join([], lastsep=' and ') == ''
    assert word_join([1], lastsep=' and ') == '1'
    assert word_join([1, 2], lastsep=' and ') == '1 and 2'
    assert word_join([1, 2, 3], lastsep=' and ') == '1, 2 and 3'
    assert word_join([1, 2, 3, 4], lastsep=' and ') == '1, 2, 3 and 4'


def test_quoted():
    assert word_join([], quoter=single) == ""
    assert word_join([1], quoter=single) == "'1'"
    assert word_join([1, 2], quoter=single) == "'1' and '2'"
    assert word_join([1, 2, 3], quoter=single) == "'1', '2', and '3'"


def test_listy():
    assert join([], quoter=single, endcaps=brackets) == "[]"
    assert join([1], quoter=single, endcaps=brackets) == "['1']"
    assert join([1, 2], quoter=single, endcaps=brackets) == "['1', '2']"
    assert join([1, 2, 3], quoter=single, endcaps=brackets) == "['1', '2', '3']"


def test_sep():
    assert join([], sep='|') == ''
    assert join([1], sep='|') == '1'
    assert join([1, 2], sep='|') == '1|2'
    assert join([1, 2, 3], sep='|') == '1|2|3'


def test_twosep_and_lastsep():
    assert join([1, 2, 3, 4], sep='|', lastsep='+') == '1|2|3+4'

    assert join([], sep='|', twosep='*', lastsep='+') == ''
    assert join([1], sep='|', twosep='*', lastsep='+') == '1'
    assert join([1, 2], sep='|', twosep='*', lastsep='+') == '1*2'
    assert join([1, 2, 3], sep='|', twosep='*', lastsep='+') == '1|2+3'
    assert join([1, 2, 3, 4], sep='|', twosep='*', lastsep='+') == '1|2|3+4'


def test_no_twostep():
    assert join([], sep='|', twosep=None, lastsep='+') == ''
    assert join([1], sep='|', twosep=None, lastsep='+') == '1'
    assert join([1, 2], sep='|', twosep=None, lastsep='+') == '1+2'
    assert join([1, 2, 3], sep='|', twosep=None, lastsep='+') == '1|2+3'
    assert join([1, 2, 3, 4], sep='|', twosep=None, lastsep='+') == '1|2|3+4'


def test_concat():
    assert concat([4, 5, 6]) == '456'
    assert concat(range(3)) == '012'
    assert concat([4, 5, 6]) == '456'
    assert concat(['a','b','c']) == 'abc'


def test_joinlines():
    assert joinlines([]) == '\n'
    assert joinlines(list('a')) == 'a\n'
    assert joinlines(list('ab')) == 'a\nb\n'
    assert joinlines(list('abc')) == 'a\nb\nc\n'


def test_and_join():
    assert and_join([]) == ''
    assert and_join([1]) == '1'
    assert and_join([1, 2]) == '1 and 2'
    assert and_join([1, 2, 3]) == '1, 2, and 3'
    assert and_join([1, 2, 3, 4]) == '1, 2, 3, and 4'


def test_or_join():
    assert or_join([]) == ''
    assert or_join([1]) == '1'
    assert or_join([1, 2]) == '1 or 2'
    assert or_join([1, 2, 3]) == '1, 2, or 3'
    assert or_join([1, 2, 3, 4]) == '1, 2, 3, or 4'


@pytest.mark.xfail(reason='road out')
def test_named_endpoints():
    mylist = list('ABCD')
    assert join(mylist, style="and") == and_join(mylist)
    assert join(mylist, style="or") == or_join(mylist)
    assert join.lines(mylist) == join(mylist, style="lines") == joinlines(mylist)


def test_items():
    alist = [1, 2, 3, 'string']
    assert items(alist) == "0: 1\n1: 2\n2: 3\n3: 'string'"

    assert items(alist, header="---") == "---\n0: 1\n1: 2\n2: 3\n3: 'string'"
    assert items(alist, footer="===") == "0: 1\n1: 2\n2: 3\n3: 'string'\n==="
    assert items(alist, header='---', footer="===") == "---\n0: 1\n1: 2\n2: 3\n3: 'string'\n==="

    try:
        from collections import OrderedDict
        answer = "this: something\nthat: else\nplus: additionally"
    except ImportError:
        OrderedDict = dict  # py26
        parts = ["this: something", "plus: additionally", "that: else"]
        # output in different order because unordred dict

    od = OrderedDict()
    od['this'] = 'something'
    od['that'] = 'else'
    od['plus'] = 'additionally'

    result1 = items(od, fmt="{key}: {value}")
    result2 = items(od, fmt="{key}: {value}", header="KEY: VALUE")

    if OrderedDict is dict:
        for p in parts:
            assert p in result1
            assert p in result2
    else:
        assert result1 == answer
        assert result2 == "KEY: VALUE\n" + answer



"""
Module to assist in the super-common operation of
wrapping values with prefix and suffix strings.
"""

import re
import six
from options import Options, OptionsClass, Prohibited, Transient
from .util import *


class BadStyleName(ValueError):
    pass


class Quoter(OptionsClass):

    """
    A quote style. Instantiate it with the style information. Call
    it with a value to quote the value.
    """

    styles = {}         # remember named styles

    options = Options(
        prefix       = None,
        suffix       = None,
        name         = None,
        margin       = 0,
        padding      = 0,
        encoding     = None,
        style        = None,
    )

    def __init__(self, *args, **kwargs):
        """
        Create a quoting style.
        """
        opts = self.options = self.__class__.options.push(kwargs)
        self._flatargs(args)
        self._register_name(opts.name)

    def _flatargs(self, args):
        """
        Consume 'flat' *args if present when object is constructed.
        """
        if args:
            opts = self.options
            used = opts.addflat(args, ['prefix', 'suffix'])
            if 'suffix' not in used:
                opts.suffix = opts.prefix
                # this suffix = prefix behavior appropriate for flat args only

    def _register_name(self, name, cls=None):
        """
        Given a name or space-separated list of names, define styles
        based on that name, and define class attributes for that style
        name as well (as long as such a style name doesn't start with
        an underscore, which is hereby prohibuted for style names to
        avoid conflict with existing Quoter methods.)
        """
        if not name:
            return
        cls = cls or getattr(self, '__class__')
        if ' ' in name:
            names = name.split()
            self.options.name = name.replace(' ', '-')
        else:
            names = [name]

        sdict = getattr(cls, 'styles')
        for n in names:
            if not n.startswith('_'):
                sdict[name] = self
                setattr(cls, n, self)
            else:
                msg = 'Style names should not start with an underscore'
                raise BadStyleName(msg)

    def _whitespace(self, opts):
        """
        Compute the appropriate margin and padding strings.
        """
        pstr = ' ' * opts.padding if isinstance(opts.padding, int) else opts.padding
        mstr = ' ' * opts.margin  if isinstance(opts.margin, int)  else opts.margin
        return (pstr, mstr)

        # could extend the padding and margins with tuples to enable
        # asymmetric before/after settings

    def _output(self, parts, opts):
        """
        Given a list of string parts, concatentate them and output
        with the given encoding (if any).
        """
        outstr = ''.join(parts)
        return outstr.encode(opts.encoding) if opts.encoding else outstr

    def __call__(self, value, **kwargs):
        """
        Quote the value, with the given padding, margin, and encoding.
        """
        opts = self.options.push(kwargs)
        if opts.style:
            cls = self.__class__
            return cls.styles[opts.style](value, **kwargs)
        else:
            pstr, mstr = self._whitespace(opts)
            suffix = opts.suffix if opts.suffix is not None else opts.prefix
            parts = [ mstr, opts.prefix, pstr, stringify(value), pstr, suffix, mstr ]
            return self._output(parts, opts)

    def clone(self, **kwargs):
        """
        Create a new instance whose options are chained to this instance's
        options (and thence to self.__class__.options). kwargs become the
        cloned instance's overlay options.
        """
        cloned = self.__class__()
        name = kwargs.get('name')
        cloned.options = self.options.push(kwargs)
        if name:
            cloned._register_name(name)
        return cloned


    but = clone


    def __getattribute__(self, name):
        if name.startswith('_') or name == 'options' or name == 'styles' \
            or name == 'set' or name =='settings' or name == 'clone' \
            or name == 'but':
            return object.__getattribute__(self, name)
        cls = object.__getattribute__(self, '__class__')
        cdict = object.__getattribute__(cls, '__dict__')
        if name in cdict:
            return cdict[name]
        return cls(name, name=name)

    # Having an auto-instantiate capability through __getattribute__ is
    # great from an ease point of view, but makes the implementation
    # very tricky to get right, especially when inheritance is
    # involved.


    # should clone take same flat args as normal instanitation?
    # currently only takes kwargs

# create some default named styles (ASCII)

quote    = Quoter("'",      name= 'default')

braces   = Quoter('{', '}', name='braces')
brackets = Quoter('[', ']', name='brackets')
angles   = Quoter('<', '>', name='angles')
parens   = Quoter('(', ')', name='parens')
qs = single   = Quoter("'",      name='single qs')
qd = double   = Quoter('"',      name='double qd')
qt = triple   = Quoter('"""',    name='triple qt')
qb = backticks= Quoter("`",      name='backticks qb')
qdb = doublebackticks= Quoter("``",      name='doublebackticks qdb')

# and some Unicode styles
anglequote = guillemet = Quoter(six.u('\u00ab'), six.u('\u00bb'), name='anglequote guillemet')
curlysingle = Quoter(six.u('\u2018'), six.u('\u2019'), name='curlysingle')
curlydouble = Quoter(six.u('\u201c'), six.u('\u201d'), name='curlydouble')

# consider adding the html entity names as the names of the above quotes
# at least as aliases

# fix issue with aliased style names


class LambdaQuoter(Quoter):

    """
    A Quoter that uses code to decide what quotes to use, based on the value.
    """

    styles = {}

    options = Quoter.options.add(
        func   = None,
        prefix = Prohibited,
        suffix = Prohibited,
    )

    def _flatargs(self, args):
        """
        Consume 'flat' *args if present when object is constructed.
        """
        if args:
            self.options.addflat(args, ['func'])

    def __call__(self, value, **kwargs):
        """
        Quote the value, based on the instance's function.
        """
        opts = self.options.push(kwargs)
        if opts.style:
            cls = self.__class__
            return cls.styles[opts.style](value, **kwargs)
        else:
            pstr, mstr = self._whitespace(opts)
            prefix, value, suffix = opts.func(value)
            parts = [mstr, prefix, pstr, stringify(value), pstr, suffix, mstr]
            return self._output(parts, opts)

lambdaq = LambdaQuoter(lambda v: ('', 'ALL YOUR BASE ARE BELONG TO US', ''))


class XMLQuoter(Quoter):

    """
    A more sophisticated quoter for XML elements that manages tags,
    namespaces, and the idea that some elements may not have contents.
    """

    styles = {}         # remember named styles

    options = Quoter.options.add(
        tag      = None,
        ns       = None,
        atts     = {},
        attquote = single,
        void     = False,
        prefix   = Prohibited,
        suffix   = Prohibited,
    )

    def __init__(self, *args, **kwargs):
        """
        Create an XMLQuoter
        """
        # Restating basic init to avoid errors of self.__getattribute__
        # that can flummox superclass instantiation
        Quoter.__init__(self)

        # take the atts kwargs for special interpretation
        kw_atts_atts = style_attribs(kwargs.pop('atts', None))

        # update remaining kwargs
        opts = self.options = self.__class__.options.push(kwargs)
        self._register_name(opts.name)

        # process flat args
        tagspec, attspec = pad(args, 2)

        # combine style attributes from kwargs and flat args
        # (in reverse order of priority)
        atts = {}
        update_style_dict(atts, kw_atts_atts)
        update_style_dict(atts, kwargs)
        update_style_dict(atts, style_attribs(attspec))
        update_style_dict(atts, style_attribs(tagspec))

        #print "args:", repr(args)
        #print "tagspec, attspec:", repr(tagspec), repr(attspec)
        #print "atts:", atts

        # finally set the object's key values
        opts.tag = atts.pop('_tag', None) or opts.tag
        opts.atts = atts


    def _attstr(self, atts, opts):
        """
        Format an attribute dict. Uses this object's default attribute quoter.
        """
        return ' '.join([''] + ["{0}={1}".format(k, opts.attquote(v)) for k, v in atts.items()])


    def __call__(self, *args, **kwargs):
        """
        Quote a value in X/HTML style, with optional attributes.
        """
        stylename = kwargs.pop('style', None)
        if stylename:
            cls = self.__class__
            return cls.styles[stylename](*args, **kwargs)
        else:
            # if not args and not opts.void:
            #    return self.clone(**kwargs)

            # take the atts kwargs for special interpretation
            kw_atts_atts = style_attribs(kwargs.pop('atts', None))

            # update remaining kwargs
            opts = self.options.push(kwargs)

            # process flat args
            if opts.void:
                spec = args[0] if args else ''
                value = None
            else:
                value, spec = pad(args, 2)

            # combine style attributes from opts, kwargs, and flat args
            # (in reverse order of priority)
            atts = {}
            update_style_dict(atts, style_attribs(opts.atts))
            update_style_dict(atts, kwargs)
            update_style_dict(atts, kw_atts_atts)
            update_style_dict(atts, style_attribs(spec))

            # if there is a local tag, let it come in force
            opts.tag = atts.pop('_tag', None) or opts.tag


            # construct the resulting attribute string
            astr = self._attstr(atts, opts) if atts else ''


            pstr, mstr = self._whitespace(opts)
            ns = opts.ns + ':' if opts.ns else ''
            if opts.void or not args:
                parts = [ mstr, '<', ns, opts.tag, astr, '>', mstr ]
            else:
                parts = [ mstr, '<', ns, opts.tag, astr, '>', pstr,
                          stringify(value),
                          pstr, '</', ns, opts.tag, '>', mstr ]
            return self._output(parts, opts)


    # could improve kwargs handling of HTMLQuoter

    # question is, should call attributes overwrite, or add to, object atts?
    # may not be a single answer - eg, in case of class especially

    # This might be case where replace is the primary option, but there's
    # an option to add (or even subtract) - say using a class Add, Plus, Subtract,
    # Minus, Relative, Rel, Delta, etc as an indicator

    # To be a full production XML quoter, might need a slightly more robust way
    # to name XML styles that include namespace names, including some sort of
    # rules for handling hyphens in the names (which cannot be in Python
    # identifiers), and perhaps for understanding namespaces (with terminating
    # colon) as part of tag specification. When a tag is auto-instantiated, it
    # could perhaps have its ns defined as part of its definition, like tag is.


class HTMLQuoter(XMLQuoter):

    """
    A more sophisticated quoter that supports attributes and void elements for HTML.
    """

    styles = {}         # remember named styles

    options = XMLQuoter.options.add(
        ns       = Prohibited,
    )

    def __init__(self, *args, **kwargs):
        XMLQuoter.__init__(self, *args, **kwargs)


# HTML/XML comments need normal, not tag-based, quoters
_markup_comment = Quoter(prefix='<!--', suffix='-->', padding=1)

html = HTMLQuoter('html')
setattr(HTMLQuoter, 'comment', _markup_comment)
HTMLQuoter.styles['comment'] = _markup_comment

# Tags that don't take content. Their payload is specified in their
# tag name and attributes, if any.
_SELFCLOSING = """
    br img hr input link meta area base col command embed keygen
    param source track
""".strip().split()

for t in _SELFCLOSING:
    hq = HTMLQuoter(t, void=True)
    setattr(HTMLQuoter, t, hq)
    XMLQuoter.styles[t] = hq


xml = XMLQuoter('xml')
setattr(XMLQuoter, 'comment', _markup_comment)
XMLQuoter.styles['comment'] = _markup_comment


# Eventually working way toward a CSS box model style formatting in which there
# can be a marginleft, marginright, paddingleft, and paddingright (i.e.
# separating left and right magin/padding specs). It might even be possible
# to provide borders (top and bottom), and to reconsider prefix and suffix
# as left and right borders. Alignment of content within a cell and various
# forms of multi-line justification might also be feasible.

# With named styles being stored in the Quoter/subclass __dict__, it is
# unclear there is any benefit to having a separate styles class attribute.
# Removing it may take some extra testing changes, but would be a simplifying
# cleanup.

# Consider abstracting the "StyleSet" functionality of a group of named
# styles from the Quoter functionality. Possibly a sidecar to the the
# ``options`` module.





from quoter.util import *
import six

def test_is_string():
    assert is_string("")
    assert is_string("This")
    assert is_string(six.u("this"))
    assert stringify(six.u("a\u2014b"))

    assert not is_string(1)
    assert not is_string(None)
    assert not is_string([1, 2, 3])
    assert not is_string(['a', 'b', 'c'])


def test_stringify():
    assert stringify('this') == 'this'
    assert stringify(4) == '4'
    assert stringify(six.u("\u2014")) == six.u("\u2014")


def test_blanknone():
    assert blanknone(None) == ''

    values = [0, 1, "", "this", True, False, list(range(4))]
    for v in values:
        assert blanknone(v) == v


def test_pad():
    assert pad([], 3) == [None, None, None]
    assert pad([1], 3) == [1, None, None]
    assert pad([1, 2], 3) == [1, 2, None]

    assert pad([], 3, -1) == [-1, -1, -1]
    assert pad([1], 3, -1) == [1, -1, -1]
    assert pad([1, 2], 3, -1) == [1, 2, -1]

    assert pad((), 3) == (None, None, None)
    assert pad((1,), 3) == (1, None, None)
    assert pad((1, 2), 3) == (1, 2, None)


def test_parse_css_selector():

    pcs = parse_css_selector

    assert pcs(None) == {}
    assert pcs("b") == { '_tag': 'b' }
    assert pcs("b.one") == { '_tag': 'b', 'class': 'one' }
    assert pcs("b.one.two") == { '_tag': 'b', 'class': 'one two' }
    assert pcs("b#name") == { '_tag': 'b', 'id': 'name' }
    assert pcs("b.one#name") == { '_tag': 'b', 'class': 'one', 'id': 'name' }
    assert pcs("b.one[andy=pandy]#name") == { '_tag': 'b', 'class': 'one',
                                              'id': 'name', 'andy': 'pandy' }


def test_combine_classes():


    cc = combine_classes
    assert cc(None, None) == None
    assert cc('one', None) == 'one'
    assert cc(None, 'two') == 'two'
    assert cc('one', 'two') == 'one two'
    assert cc('two', 'one') == 'two one'
    assert cc(None, 'billy idol') == 'billy idol'
    assert cc('billy jones', 'billy idol') == 'billy jones idol'
    assert cc('billy jones', 'idol billy billy nix') == 'billy jones idol nix'

    # simulate the reverse_classes setting
    cc = lambda x,y: combine_classes(y, x)
    assert cc(None, None) == None
    assert cc('one', None) == 'one'
    assert cc(None, 'two') == 'two'
    assert cc('one', 'two') == 'two one'
    assert cc('two', 'one') == 'one two'
    assert cc(None, 'billy idol') == 'billy idol'
    assert cc('billy jones', 'billy idol') == 'billy idol jones'
    assert cc('billy jones', 'idol billy billy nix') == 'idol billy nix jones'


def test_update_style_dict():
    d = { 'id': 'one' }
    update_style_dict(d, { 'class': 'billy idol'})
    assert d == { 'id': 'one', 'class': 'billy idol'}

    # first with simpler, non-class-reversing behavior
    d = { 'id': 'one', 'class': 'billy jones' }
    update_style_dict(d, { 'class': 'billy idol'}, reverse_class=False)
    assert d == { 'id': 'one', 'class': 'billy jones idol'}

    d = { 'id': 'one', 'class': 'billy jones' }
    update_style_dict(d, { 'class': 'idol billy billy nix',
                            'src': 'file.jpg' },
                      reverse_class=False)
    assert d == { 'id': 'one', 'class': 'billy jones idol nix',
                   'src': 'file.jpg' }


    # then with normal class-reversing behavior
    d = { 'id': 'one', 'class': 'billy jones' }
    update_style_dict(d, { 'class': 'billy idol'})
    assert d == { 'id': 'one', 'class': 'billy idol jones'}

    d = { 'id': 'one', 'class': 'billy jones' }
    update_style_dict(d, { 'class': 'idol billy billy nix',
                            'src': 'file.jpg' })
    assert d == { 'id': 'one', 'class': 'idol billy nix jones',
                   'src': 'file.jpg' }



#!/usr/bin/env python

from setuptools import setup
from codecs import open
import sys

_PY26 = sys.version_info[:2] == (2, 6)
STUF = 'stuf==0.9.14' if _PY26 else 'stuf>=0.9.16'


def lines(text):
    """
    Returns each non-blank line in text enclosed in a list.
    See http://pypi.python.org/pypi/textdata for more sophisticated version.
    """
    return [l.strip() for l in text.strip().splitlines() if l.strip()]


setup(
    name='quoter',
    version='1.5.1',
    author='Jonathan Eunice',
    author_email='jonathan.eunice@gmail.com',
    description="Powerful way to construct text, HTML, and XML, plus a kick-ass join",
    long_description=open('README.rst', encoding='utf-8').read(),
    url='https://bitbucket.org/jeunice/quoter',
    license='Apache License 2.0',
    packages=['quoter'],
    setup_requires=[],
    install_requires=['six>=1.9', STUF, 'options>=1.2.2'],
    tests_require=['tox', 'pytest', 'pytest-cov', 'coverage', 'six>=1.9'],
    test_suite="test",
    zip_safe=False,  # it really is, but this will prevent weirdness
    keywords='quote wrap prefix suffix endcap repr representation html xml join',
    classifiers=lines("""
        Development Status :: 4 - Beta
        Operating System :: OS Independent
        License :: OSI Approved :: Apache Software License
        Intended Audience :: Developers
        Programming Language :: Python
        Programming Language :: Python :: 2
        Programming Language :: Python :: 2.7
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.2
        Programming Language :: Python :: 3.3
        Programming Language :: Python :: 3.4
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: Implementation :: CPython
        Programming Language :: Python :: Implementation :: PyPy
        Topic :: Software Development :: Libraries :: Python Modules
        Topic :: Text Processing :: Filters
        Topic :: Text Processing :: Markup
        Topic :: Text Processing :: Markup :: HTML
        Topic :: Text Processing :: Markup :: XML
    """)
)



from quoter import *
import pytest


def test_braces():
    assert braces('this') == '{this}'
    assert braces('this', padding=1) == '{ this }'
    assert braces('this', margin=1) == ' {this} '
    assert braces('this', padding=1, margin=1) == ' { this } '


def test_brackets():
    assert brackets('this') == '[this]'
    assert brackets('this', padding=1) == '[ this ]'
    assert brackets('this', margin=1) == ' [this] '
    assert brackets('this', padding=1, margin=1) == ' [ this ] '


def test_clone():
    # make sure clones properly registered just like normally defined
    # styles, that clones are different from original, that they yield
    # the expected results, and that after cloning, the originals
    # behave exactly as before

    bar2 = brackets.clone(margin=2, name='bar2')
    assert bar2 is not brackets
    assert bar2('this') == '  [this]  '
    assert brackets('this') == '[this]'
    assert quote.bar2('this') == '  [this]  '

    bb = html.b.clone(padding=1, margin=2)
    assert bb is not html.b
    assert 'bb' not in HTMLQuoter.styles
    assert bb('this') == '  <b> this </b>  '
    assert html.b('this') == '<b>this</b>'

    sp = html.b.but(atts='span')
    assert sp is not html.b
    assert sp('something') == '<span>something</span>'
    assert sp('else', '.other') == "<span class='other'>else</span>"


def test_but():
    """
    Test the but alias for clone.
    """
    bar2 = brackets.but(margin=2)
    assert bar2 is not brackets
    assert bar2('this') == '  [this]  '
    assert brackets('this') == '[this]'

    bb = html.b.but(padding=1, margin=2)
    assert bb is not html.b
    assert bb('this') == '  <b> this </b>  '
    assert html.b('this') == '<b>this</b>'


def test_set():
    bar2 = brackets.clone(margin=2, name='bar2')
    assert bar2('this') == '  [this]  '
    assert quote.bar2('this') == '  [this]  '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

    # check setting options
    bar2.set(margin=1, padding=1)
    assert bar2('this') == ' [ this ] '
    assert quote.bar2('this') == ' [ this ] '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

    # check setting options
    bar2.set(prefix='||', padding=1)
    assert bar2('this') == ' || this ] '
    assert quote.bar2('this') == ' || this ] '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

def test_set_example():
    bars = Quoter('|')
    assert bars('x') == '|x|'
    bars.set(prefix='||', suffix='||')
    assert bars('x') == '||x||'
    bars.set(padding=1)
    assert bars('x') == '|| x ||'

    bart = bars.clone(prefix=']', suffix = '[')
    assert bart('x') == '] x ['


# NB There is an error in namespace managment.
# HTML attributes should be limited to HTMLQuoter,
# but there is currently leakage. Styles seem to be
# defined in Quoter, not the most proximate superclass


@pytest.mark.skipif(True, reason='road closed')
def test_chars():
    percent = Quoter(chars='%%')
    assert percent('something') == '%something%'
    doublea = Quoter(chars='<<>>')
    assert doublea('AAA') == '<<AAA>>'

# From a quoter point of view, the chars kwarg is totally gilding the
# lily, and unnecessary. But from the point of view of exploring and
# exercising the underlying options package, it is quite interesting,
# in that it requires a mapping of one user-level argument into two
# different underlying parameters. It's a good edge-case for testing.


def test_auto_stringification():
    assert brackets(12) == '[12]'
    assert braces(4.4) == '{4.4}'
    assert double(None) == '"None"'
    assert single(False) == "'False'"


def test_shortcuts():
    assert ' '.join([qs('one'), qd('two'), qt('three'), qb('and'), qdb('four')]) == \
        "'one' \"two\" \"\"\"three\"\"\" `and` ``four``"


def test_instant():
    assert Quoter('+[ ', ' ]+')('castle') == '+[ castle ]+'


def test_lambda():
    f = lambda v: ('(', abs(v), ')') if v < 0 else ('', v, '')
    financial = LambdaQuoter(f)
    assert financial(-10) == '(10)'
    assert financial(44) == '44'

    password = LambdaQuoter(lambda v: ('', 'x' * len(v), ''))
    assert password('secret!') == 'xxxxxxx'

    wf = lambda v:  ('**', v, '**') if v < 0 else ('', v, '')
    warning = LambdaQuoter(wf, name='warning')
    assert warning(12) == '12'
    assert warning(-99) == '**-99**'
    assert warning(-99, padding=1) == '** -99 **'

    assert lambdaq.warning(12) == '12'
    assert lambdaq.warning(-99) == '**-99**'
    assert lambdaq.warning(-99, padding=1) == '** -99 **'


def test_lambdaq_named_style():
    assert lambdaq(44, style='warning') == '44'
    assert lambdaq(-44, style='warning') == '**-44**'


def test_html_examples():
    assert html.p("A para", ".focus") == "<p class='focus'>A para</p>"
    assert html.img('.large', src='file.jpg') in [
        "<img class='large' src='file.jpg'>",
        "<img src='file.jpg' class='large'>"
        ]
    assert html.br() == "<br>"
    assert html.comment("content ends here") == "<!-- content ends here -->"


    assert html('hey', 'p#one.main.special[lang=en]') in [
        "<p id='one' class='main special' lang='en'>hey</p>",
        "<p id='one' lang='en' class='main special'>hey</p>",
        "<p class='main special' id='one' lang='en'>hey</p>",
        "<p class='main special' lang='en' id='one'>hey</p>",
        "<p lang='en' id='one' class='main special'>hey</p>",
        "<p lang='en' class='main special' id='one'>hey</p>",
        ]
        # all the permutations!


def test_examples():
    assert single('this') == "'this'"
    assert double('that') == '"that"'
    assert backticks('ls -l') == "`ls -l`"
    assert braces('curlycue') == "{curlycue}"
    assert braces('curlysue', padding=1) == '{ curlysue }'

    bars = Quoter('|')
    assert bars('x') == '|x|'

    plus = Quoter('+', '')
    assert plus('x') == '+x'

    variable = Quoter('${', '}', name='variable')
    assert variable('x') == '${x}'


def test_attribute_invocations():
    assert single('something') == quote.single('something')
    assert single('something', margin=2, padding=3) == quote.single('something', margin=2, padding=3)
    assert braces('b') == quote.braces('b')

    # now test wholesale
    names = 'braces brackets angles parens qs qd qt qb single double triple ' +\
            'backticks anglequote guillemet curlysingle curlydouble'
    for name in names.split():
        main = eval(name)
        attr = eval('quote.' + name)
        assert main is attr
        assert main('string') == attr('string')


def test_quote_shortcut():
    variable = Quoter('${', '}', name='variable')
    assert variable('x') == '${x}'

    assert quote('myvar', style='variable') == '${myvar}'

    assert quote('this', style='braces') == '{this}'


def test_named_quote_style():
    assert quote('this', style='variable') == '${this}'
    assert braces('this', style='variable') == '${this}'


def test_redef():
    braces = Quoter('{', '}', padding=1, name='braces')
    assert braces('this') == '{ this }'
    assert braces('this', padding=0) == '{this}'


def test_para():
    para = HTMLQuoter('p')
    # assert para('this is great!', {'class':'emphatic'}) == "<p class='emphatic'>this is great!</p>"
    assert para('this is great!', '.emphatic') == "<p class='emphatic'>this is great!</p>"
    assert para('First para!', '#first') == "<p id='first'>First para!</p>"
    assert para('First para!', '#first', atts='.one') in [
        "<p id='first' class='one'>First para!</p>",
        "<p class='one' id='first'>First para!</p>"]

    para_e = HTMLQuoter('p.emphatic')
    assert para_e('this is great!') == "<p class='emphatic'>this is great!</p>"
    assert para_e('this is great?', '.question') == "<p class='question emphatic'>this is great?</p>"

    para = HTMLQuoter('p', attquote=double)
    assert para('this is great!', {'class':'emphatic'}) == '<p class="emphatic">this is great!</p>'

    div = HTMLQuoter('div', attquote=double)
    assert div('something', '.todo') == '<div class="todo">something</div>'


def test_css_selector():
    assert html('joe', 'b.name') == "<b class='name'>joe</b>"
    assert xml('joe', 'b.name') == "<b class='name'>joe</b>"

    assert xml('joe', 'name#emp0193') == "<name id='emp0193'>joe</name>"


def test_void():
    br = HTMLQuoter('br', void=True)
    assert br() == '<br>'

    img = HTMLQuoter('img', void=True)
    assert img() == '<img>'
    assert img(src="this") == "<img src='this'>"
    assert img('.roger', src="this") == "<img class='roger' src='this'>" or \
           img('.roger', src="this") == "<img src='this' class='roger'>"


def test_xml_examples():
    item = XMLQuoter(tag='item', ns='inv', name='item inv_item')
    assert item('an item') == '<inv:item>an item</inv:item>'
    assert xml.item('another') == '<inv:item>another</inv:item>'
    assert xml.inv_item('yet another') == '<inv:item>yet another</inv:item>'
    assert xml.thing('something') == '<thing>something</thing>'
    assert xml.special('else entirely', '#unique') == \
            "<special id='unique'>else entirely</special>"


def test_xml_auto_and_attributes():

    assert xml.root('this') == '<root>this</root>'
    assert xml.root('this', ns='one') == '<one:root>this</one:root>'
    assert xml.branch('that') == '<branch>that</branch>'
    assert xml.branch('that', ns='two') == '<two:branch>that</two:branch>'

    assert xml.comment('hidden') == '<!-- hidden -->'
    assert xml.comment('hidden', padding=0) == '<!--hidden-->'


def test_html_auto_and_attributes():
    assert html.b('bold') == '<b>bold</b>'
    assert html.emphasis('bold') == '<emphasis>bold</emphasis>'
    assert html.strong('bold') == '<strong>bold</strong>'
    assert html.strong('bold', padding=1) == '<strong> bold </strong>'
    assert html.strong('bold', margin=1) == ' <strong>bold</strong> '

    assert html.comment('XYZ') == '<!-- XYZ -->'
    assert html.comment('XYZ', padding=0) == '<!--XYZ-->'

    assert html.br() == '<br>'
    assert html.img(src='one') == "<img src='one'>"


def test_named_styles_in_proper_homes():
    assert 'x' not in Quoter.styles
    assert 'x' not in HTMLQuoter.styles
    assert 'x' not in XMLQuoter.styles

    x = Quoter('X', name='x')
    assert x('y') == 'XyX'
    assert 'x' in Quoter.styles

    assert html.x("y") == "<x>y</x>"
    assert 'x' in HTMLQuoter.styles

    assert xml.x("y") == "<x>y</x>"
    assert 'x' in XMLQuoter.styles


def test_named_xml_and_html_styles():
    XMLQuoter('book', name='book')
    assert xml("this", style="book") == "<book>this</book>"


def test_bad_style_names():
    with pytest.raises(BadStyleName):
        v = Quoter('v', name="_v")


# some problem exists with explicit atts setting here
@pytest.mark.skipif('True')
def test_xml_autogenerate():
    more = xml.b.clone(atts='.this')
    assert more('x') == "<b class='this'>x</b>"
